//////////////////////////////////////////////
// PastCode.h
//
// Some code from last tutorials used
// to initialize and clean a bit the main
// file
//////////////////////////////////////////////

/// Includes ---------------------------------

// nkLog
#include <NilkinsLog/ConsoleLogger.h>

// nkGraphics
#include <NilkinsGraphics/Buffers/Buffer.h>
#include <NilkinsGraphics/Buffers/BufferManager.h>

#include <NilkinsGraphics/Cameras/Camera.h>
#include <NilkinsGraphics/Cameras/CameraManager.h>

#include <NilkinsGraphics/Log/LogManager.h>

#include <NilkinsGraphics/Graph/Entity.h>
#include <NilkinsGraphics/Graph/SubEntity.h>

#include <NilkinsGraphics/Graph/Node.h>
#include <NilkinsGraphics/Graph/NodeManager.h>

#include <NilkinsGraphics/Meshes/Load/MeshLoader.h>

#include <NilkinsGraphics/Meshes/Mesh.h>
#include <NilkinsGraphics/Meshes/MeshManager.h>

#include <NilkinsGraphics/Programs/Program.h>
#include <NilkinsGraphics/Programs/ProgramManager.h>
#include <NilkinsGraphics/Programs/ProgramSourcesHolder.h>

#include <NilkinsGraphics/RenderContexts/RenderContextManager.h>

#include <NilkinsGraphics/RenderQueues/RenderQueue.h>
#include <NilkinsGraphics/RenderQueues/RenderQueueManager.h>

#include <NilkinsGraphics/Samplers/Sampler.h>
#include <NilkinsGraphics/Samplers/SamplerManager.h>

#include <NilkinsGraphics/Shaders/Memory/ConstantBuffer.h>
#include <NilkinsGraphics/Shaders/Memory/ShaderInstanceMemorySlot.h>
#include <NilkinsGraphics/Shaders/Memory/ShaderPassMemorySlot.h>

#include <NilkinsGraphics/Shaders/Shader.h>
#include <NilkinsGraphics/Shaders/ShaderManager.h>

#include <NilkinsGraphics/Systems/MainSystem.h>

#include <NilkinsGraphics/Textures/Texture.h>
#include <NilkinsGraphics/Textures/TextureManager.h>

// nkResources
#include <NilkinsResources/ResourceManager.h>

// nkWinUi
#include <NilkinsWinUi/Systems/MainSystem.h>

// Standards
#include <chrono>

#include <memory>

/// Internals : Buffers ----------------------

void prepareRaytracedBuffer ()
{
	nkGraphics::Buffer* buffer = nkGraphics::BufferManager::getInstance()->createOrRetrieve("raytracedBuffer") ;
	buffer->prepareForComputeResourceUsage() ;
	buffer->prepareForShaderResourceUsage() ;

	buffer->setElementByteSize(4 * sizeof(float)) ;
	buffer->setElementCount(800 * 600) ;

	buffer->load() ;
}

/// Internals : Mesh -------------------------

void prepareMeshFromFile ()
{
	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("sphere") ;

	nkGraphics::MeshLoader::fillBufferFromFile(mesh, "sphere.obj") ;
	mesh->load() ;
}

void addMeshToRenderQueue ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;

	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("sphere") ;

	nkGraphics::Entity* ent = rq->addEntity() ;
	nkGraphics::SubEntity* subEnt = ent->addChild() ;

	subEnt->setMesh(mesh) ;
}

void addEntityToGraph ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;
	nkGraphics::Entity* ent = rq->getEntity(0) ;

	nkGraphics::Node* node = nkGraphics::NodeManager::getInstance()->create() ;
	node->setPositionAbsolute(nkMaths::Vector(0.f, 0.f, 10.f)) ;

	ent->setParentNode(node) ;
}

/// Internals : Texture ----------------------

void prepareTexture ()
{
	nkGraphics::Texture* tex = nkGraphics::TextureManager::getInstance()->createOrRetrieve("tex") ;

	tex->setResourcePath("yokohamaNight.dds") ;
	tex->setGammaCorrected(false) ;
	tex->load() ;
}

void prepareSampler ()
{
	nkGraphics::Sampler* sampler = nkGraphics::SamplerManager::getInstance()->createOrRetrieve("sampler") ;
	sampler->load() ;
}

/// Internals : Shader -----------------------

void prepareEnvProgram ()
{
	// This program will be used to render a background in sync with the sphere rendered
	nkGraphics::Program* envProgram = nkGraphics::ProgramManager::getInstance()->createOrRetrieve("envProgram") ;
	nkGraphics::ProgramSourcesHolder envSources ;

	envSources.setVertexMemory
	(
		R"eos(
			// The camDir entry is an array of 4 float4
			// They will correspond to the 4 corners of the screen
			// This is true because we will use this shader in a post process pass
			// Such a pass map a quad to the screen and renders it using the shader provided
			cbuffer constants
			{
				float4 camDir [4] ;
			}

			// The vertexId is fed by DirectX and correspond to the index of the vertex being processed
			// A pass slot for camera direction feeds the corners in sync with the order the vertices are put in the mesh
			// As such, we can use it to directly index which direction a vertex corresponds to
			struct VertexInput
			{
				float4 position : POSITION ;
				uint vertexId : SV_VertexID ;
			} ;

			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float4 camDir : CAMDIR ;
			} ;

			PixelInput main (VertexInput input)
			{
				PixelInput result ;

				result.position = input.position ;
				result.camDir = camDir[input.vertexId] ;

				return result ;
			}
		)eos"
	) ;

	envSources.setPixelMemory
	(
		R"eos(
			// The camera direction will be interpolated between the 4 vertices, for each pixel
			// This means we will get the direction for a given pixel in the world
			// Using that knowledge we only need to sample the cube map
			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float4 camDir : CAMDIR ;
			} ;

			TextureCube envMap : register(t0) ;
			SamplerState customSampler : register(s0) ;

			float4 main (PixelInput input) : SV_TARGET
			{
				return envMap.Sample(customSampler, normalize(input.camDir)) ;
			}
		)eos"
	) ;

	envProgram->setFromMemory(envSources) ;
	envProgram->load() ;
}

void prepareBufferCopyProgram ()
{
	nkGraphics::Program* bufferCopyProgram = nkGraphics::ProgramManager::getInstance()->createOrRetrieve("bufferCopyProgram") ;
	nkGraphics::ProgramSourcesHolder sources ;

	sources.setVertexMemory
	(
		R"eos(
			struct VertexInput
			{
				float4 position : POSITION ;
				float2 uvs : TEXCOORD0 ;
			} ;

			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float2 uvs : TEXCOORD0 ;
				uint4 texInfos : TEXINFOS ;
			} ;

			cbuffer passConstants
			{
				uint4 texInfos ;
			}

			PixelInput main (VertexInput input)
			{
				PixelInput result ;

				result.position = input.position ;
				result.uvs = input.uvs ;
				result.texInfos = texInfos ;

				return result ;
			}
		)eos"
	) ;

	sources.setPixelMemory
	(
		R"eos(
			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float2 uvs : TEXCOORD0 ;
				uint4 texInfos : TEXINFOS ;
			} ;

			StructuredBuffer<float4> inputBuf : register(t0) ;

			float4 main (PixelInput input) : SV_TARGET
			{
				uint2 index = uint2(input.uvs.x * input.texInfos.x, input.uvs.y * input.texInfos.y) ;
				float4 inputColor = inputBuf[index.x + index.y * input.texInfos.x] ;

				if (inputColor.a < 0.001)
					discard ;

				return inputColor ;
			}
		)eos"
	) ;

	bufferCopyProgram->setFromMemory(sources) ;
	bufferCopyProgram->load() ;
}

void prepareEnvShader ()
{
	// Prepare the post process shader
	nkGraphics::Program* envProgram = nkGraphics::ProgramManager::getInstance()->get("envProgram") ;

	nkGraphics::Texture* tex = nkGraphics::TextureManager::getInstance()->get("tex") ;
	nkGraphics::Sampler* sampler = nkGraphics::SamplerManager::getInstance()->get("sampler") ;

	nkGraphics::Shader* envShader = nkGraphics::ShaderManager::getInstance()->createOrRetrieve("envShader") ;
	
	envShader->setAttachedShaderProgram(envProgram) ;

	nkGraphics::ConstantBuffer* cBuffer = envShader->addConstantBuffer(0) ;

	nkGraphics::ShaderPassMemorySlot* slot = cBuffer->addPassMemorySlot() ;
	slot->setAsCamCornersWorld() ;

	envShader->addTexture(tex, 0) ;
	envShader->addSampler(sampler, 0) ;

	envShader->load() ;
}

void prepareBufferCopyShader ()
{
	nkGraphics::Shader* bufferCopyShader = nkGraphics::ShaderManager::getInstance()->createOrRetrieve("bufferCopy") ;
	nkGraphics::Program* bufferCopyProgram = nkGraphics::ProgramManager::getInstance()->get("bufferCopyProgram") ;

	bufferCopyShader->setAttachedShaderProgram(bufferCopyProgram) ;

	nkGraphics::ConstantBuffer* cBuffer = bufferCopyShader->addConstantBuffer(0) ;

	nkGraphics::ShaderPassMemorySlot* slot = cBuffer->addPassMemorySlot() ;
	slot->setAsTargetSize() ;

	nkGraphics::Buffer* buffer = nkGraphics::BufferManager::getInstance()->createOrRetrieve("raytracedBuffer") ;
	bufferCopyShader->addTexture(buffer, 0) ;

	bufferCopyShader->load() ;
}

void assignShader ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;
	nkGraphics::Entity* ent = rq->getEntity(0) ;
	nkGraphics::Shader* shader = nkGraphics::ShaderManager::getInstance()->get("shader") ;

	ent->setShader(shader) ;
}

/// Internals : Base -------------------------

void baseInit ()
{
	// Prepare the mesh we will show
	prepareMeshFromFile() ;
	addMeshToRenderQueue() ;
	addEntityToGraph() ;

	// Prepare the texture we will use
	prepareTexture() ;
	prepareSampler() ;

	// Load the post processing ones
	prepareEnvProgram() ;
	prepareEnvShader() ;
	prepareBufferCopyProgram() ;
	prepareBufferCopyShader() ;
}

void renderLoop (nkGraphics::RenderContext* context)
{
	// Get some data to prepare for the rendering loop
	nkGraphics::Camera* camera = nkGraphics::CameraManager::getInstance()->getDefaultCam() ;
	nkGraphics::Node* sphereNode = nkGraphics::RenderQueueManager::getInstance()->get(0)->getEntity(1)->getParentNode() ;

	nkGraphics::MainSystem* nkGraphicsSystem = nkGraphics::MainSystem::getInstance() ;
	nkWinUi::MainSystem* nkWinUiSystem = nkWinUi::MainSystem::getInstance() ;

	const float loopTimeMs = 5 * 1000.f ;
	std::chrono::system_clock::time_point firstNow = std::chrono::system_clock::now() ;

	const nkMaths::Vector sphereCenter = nkMaths::Vector(0.f, 0.f, 10.f) ;

	// Our rendering loop is ready to start
	while (nkGraphicsSystem->getHasRunToContinue())
	{
		// Frame the graphics system
		if (!nkGraphicsSystem->frame(context))
			break ;

		// Loop the camera around the sphere
		std::chrono::system_clock::time_point now = std::chrono::system_clock::now() ;
		float currentTimeMs = std::chrono::duration_cast<std::chrono::microseconds>(now - firstNow).count() / 1000.f ;
		
		float currentLoopT = (std::fmod(currentTimeMs, loopTimeMs) / loopTimeMs) * 2.f - 1.f ;
		currentLoopT *= 3.14 ;
		
		nkMaths::Vector loopPos (std::cos(currentLoopT) * 10.f, 0.f, std::sin(currentLoopT) * 10.f) ;
		loopPos += sphereCenter ;

		camera->setPositionAbsolute(loopPos) ;
		camera->lookAt(sphereCenter, nkMaths::Vector(0.f, 1.f, 0.f)) ;

		loopPos = nkMaths::Vector(std::cos(currentLoopT) * 3.f, std::sin(currentLoopT) * 3.f, 0.f) ;
		loopPos += sphereCenter ;
		sphereNode->setPositionAbsolute(loopPos) ;

		// Tick the windowing system
		nkWinUiSystem->tick() ;
	}
}